------------------------------------------------------------------------
-- Type system for a logical framework with explicit substitutions
------------------------------------------------------------------------

-- Based on Conor's way to define a data type for an applicative
-- language (see AFP lecture notes, or "The view from the left"). His
-- method probably comes from someone else (Altenkirch and Reus?).

infixl 90 @
infixl 85 ⊙
infixl 80 ↑
infixl 70 /, /⊢
infixl 50 ▻, ▻′, ▻ʳ, ▻Cong, ▻ʳCong, ++
infixl 47 `transTm`
infixl 45 ∷≡⊢, ∷≡∋
infix  40 ∋, ⊢, ≜, ≜→, ≛, ≈⊢, ≈∋, ≘

mutual
  ----------------------------------------------------------------------
  -- Typing rules

  -- Contexts.

  data Ctxt : Set where
    ε   : Ctxt
    (▻) : (Γ : Ctxt) -> Ty Γ -> Ctxt

  -- τ : Ty Γ means that τ is a type given the context Γ.

  data Ty : (Γ : Ctxt) -> Set where
    ✶             : (Γ : Ctxt) -> Ty Γ
    Π (Γ : Ctxt)  : (τ : Ty Γ) -> Ty (Γ ▻ τ) -> Ty Γ
    El (Γ : Ctxt) : Γ ⊢ ✶ Γ -> Ty Γ

  -- Subst Γ Δ is a substitution (or weakening) of something with Γ as
  -- context into something with Δ as context.

  data Subst : Ctxt -> Ctxt -> Set where
    -- Weakening.
    wk (Γ : Ctxt) : (σ : Ty Γ) -> Subst Γ (Γ ▻ σ)
    -- Substituting a single variable.
    sub (Γ : Ctxt) (τ : Ty Γ) : Γ ⊢ τ -> Subst (Γ ▻ τ) Γ
    -- Lifting.
    (↑) (Γ, Δ : Ctxt)
      : (ρ : Subst Γ Δ) -> (σ : Ty Γ) -> Subst (Γ ▻ σ) (Δ ▻ σ / ρ)
    -- Identity.
    id : (Γ : Ctxt) -> Subst Γ Γ
    -- Composition.
    (⊙) (Γ, Δ, X : Ctxt) : Subst Γ Δ -> Subst Δ X -> Subst Γ X

  -- Γ ∋ σ represents variables of type σ in the context Γ.

  data (∋) : (Γ : Ctxt) -> Ty Γ -> Set where
    vz (Γ : Ctxt)
      : (τ : Ty Γ) -> Γ ▻ τ ∋ τ / wk τ
    vs (Γ : Ctxt) (τ : Ty Γ)
      : Γ ∋ τ -> (σ : Ty Γ) -> Γ ▻ σ ∋ τ / wk σ
    (∷≡∋) (Γ : Ctxt) (τ¹, τ² : Ty Γ)
      : Γ ∋ τ¹ -> τ¹ ≛ τ² -> Γ ∋ τ²

  -- Γ ⊢ τ represents terms of type τ in the context Γ. (Since
  -- everything is explicitly typed and we also record casts this type
  -- really represents typing derivations, not just terms.)

  data (⊢) : (Γ : Ctxt) -> Ty Γ -> Set where
    var (Γ : Ctxt) (τ : Ty Γ)
      : Γ ∋ τ -> Γ ⊢ τ
    λ (Γ : Ctxt) (τ₁ : Ty Γ) (τ₂ : Ty (Γ ▻ τ₁))
      : Γ ▻ τ₁ ⊢ τ₂ -> Γ ⊢ Π τ₁ τ₂
    (@) (Γ : Ctxt) (τ₁ : Ty Γ) (τ₂ : Ty (Γ ▻ τ₁))
      : Γ ⊢ Π τ₁ τ₂ -> (t₂ : Γ ⊢ τ₁) -> Γ ⊢ τ₂ / sub t₂
    (∷≡⊢) (Γ : Ctxt) (τ¹, τ² : Ty Γ)
      : Γ ⊢ τ¹ -> τ¹ ≛ τ² -> Γ ⊢ τ²
    (/⊢) (Γ, Δ : Ctxt) (τ : Ty Γ)
      : Γ ⊢ τ -> (ρ : Subst Γ Δ) -> Δ ⊢ τ / ρ

  ----------------------------------------------------------------------
  -- Synonyms

  -- Constructors cannot occur in types in the mutual block in which
  -- they are defined.

  (▻′) : (Γ : Ctxt) -> Ty Γ -> Ctxt
  (▻′) = (▻)

  ✶' : (Γ : Ctxt) -> Ty Γ
  ✶' = ✶

  wk' (Γ : Ctxt) : (σ : Ty Γ) -> Subst Γ (Γ ▻′ σ)
  wk' = wk

  ----------------------------------------------------------------------
  -- Substitution functions

  -- Applies a substitution to a type

  (/) (Γ, Δ : Ctxt) : Ty Γ -> Subst Γ Δ -> Ty Δ
  ✶ Γ     / ρ = ✶ _
  Π τ₁ τ₂ / ρ = Π (τ₁ / ρ) (τ₂ / ρ ↑ τ₁)
  El t    / ρ = El $ force✶Subst (t /⊢ ρ)

  force✶Subst (Γ, Δ : Ctxt) (ρ : Subst Γ Δ)
    : Δ ⊢ ✶' Γ / ρ -> Δ ⊢ ✶' Δ
  force✶Subst t = t

  init (Γ, Δ : Ctxt) (τ : Ty Γ)
    : Subst (Γ ▻′ τ) Δ -> Subst Γ Δ
  init {Γ} {Δ} {τ} ρ = wk τ ⊙ ρ

  last' (Γ, Δ : Ctxt) (τ : Ty Γ)
    : (ρ : Subst (Γ ▻′ τ) Δ) -> Δ ⊢ τ / wk' τ / ρ
  last' {Γ} {Δ} {τ} ρ = var (vz τ) /⊢ ρ

  ----------------------------------------------------------------------
  -- Equality

  -- Context equality. (The least congruence on Ctxt.)

  data (≜) : (Γ¹, Γ² : Ctxt) -> Set where

    -- Equivalence relation axioms proved below.

    εCong : ε ≜ ε
    (▻Cong) (Γ¹, Γ² : Ctxt) (τ¹ : Ty Γ¹) (τ² : Ty Γ²)
      : Γ¹ ≜ Γ² -> τ¹ ≛ τ² -> Γ¹ ▻ τ¹ ≜ Γ² ▻ τ²

  -- Type equality. (The least congruence on Ty.)

  (≛) (Γ¹, Γ² : Ctxt) : Ty Γ¹ -> Ty Γ² -> Set
  (≛) = TyEq _ _

  data TyEq : (Γ¹, Γ² : Ctxt) -> Ty Γ¹ -> Ty Γ² -> Set where

    -- Equivalence relation axioms proved below.

    ✶Cong (Γ¹, Γ² : Ctxt)
      : Γ¹ ≜ Γ² -> ✶ Γ¹ ≛ ✶ Γ²
    ΠCong (Γ¹ : Ctxt) (τ₁¹ : Ty Γ¹) (τ₂¹ : Ty (Γ¹ ▻ τ₁¹))
          (Γ² : Ctxt) (τ₁² : Ty Γ²) (τ₂² : Ty (Γ² ▻ τ₁²))
      : τ₁¹ ≛ τ₁² -> τ₂¹ ≛ τ₂² -> Π τ₁¹ τ₂¹ ≛ Π τ₁² τ₂²
    ElCong (Γ¹ : Ctxt) (t¹ : Γ¹ ⊢ ✶ Γ¹)
           (Γ² : Ctxt) (t² : Γ² ⊢ ✶ Γ²)
      : t¹ ≈⊢ t² -> El t¹ ≛ El t²

  -- Substitution equality. This is a first-order variant of the
  -- extensional equality /≘/ defined in Extensional. The higher-order
  -- variant is more natural, but is harder to control,
  -- termination-wise, especially in the current mutual block. The two
  -- equalities are proved to be equivalent in Extensional.

  (≘) (Γ¹, Γ², Δ¹, Δ² : Ctxt) : Subst Γ¹ Δ¹ -> Subst Γ² Δ² -> Set
  (≘) = SubstEq _ _ _ _

  data SubstEq :  (Γ¹, Γ², Δ¹, Δ² : Ctxt)
               -> Subst Γ¹ Δ¹ -> Subst Γ² Δ² -> Set where

    -- Congruence axioms proved below.

    emptyCong (Δ¹ : Ctxt) (ρ¹ : Subst ε Δ¹)
              (Δ² : Ctxt) (ρ² : Subst ε Δ²)
      : Δ¹ ≜ Δ² -> ρ¹ ≘ ρ²
    snocCong (Γ¹, Δ¹ : Ctxt) (τ¹ : Ty Γ¹) (ρ¹ : Subst (Γ¹ ▻ τ¹) Δ¹)
             (Γ², Δ² : Ctxt) (τ² : Ty Γ²) (ρ² : Subst (Γ² ▻ τ²) Δ²)
      : init ρ¹ ≘ init ρ² -> τ¹ ≛ τ² -> last' ρ¹ ≈⊢ last' ρ² -> ρ¹ ≘ ρ²

  -- Variable equality. (The least congruence on (∋) subject to the
  -- cast laws given below.)

  (≈∋) (Γ¹, Γ² : Ctxt) (τ¹ : Ty Γ¹) (τ² : Ty Γ²)
    : Γ¹ ∋ τ¹ -> Γ² ∋ τ² -> Set
  (≈∋) = VarEq _ _ _ _

  data VarEq :  (Γ¹, Γ² : Ctxt) -> (τ¹ : Ty Γ¹) -> (τ² : Ty Γ²)
             -> Γ¹ ∋ τ¹ -> Γ² ∋ τ² -> Set where

    -- Equivalence relation axioms proved below.

    -- Congruence (castVarCong proved below).

    vzCong (Γ¹ : Ctxt) (τ¹ : Ty Γ¹)
           (Γ² : Ctxt) (τ² : Ty Γ²)
      : τ¹ ≛ τ² -> vz τ¹ ≈∋ vz τ²
    vsCong (Γ¹ : Ctxt) (τ¹, σ¹ : Ty Γ¹) (v¹ : Γ¹ ∋ τ¹)
           (Γ² : Ctxt) (τ², σ² : Ty Γ²) (v² : Γ² ∋ τ²)
      :  v¹ ≈∋ v² -> σ¹ ≛ σ²
      -> vs v¹ σ¹ ≈∋ vs v² σ²

    -- Casts. Why have I chosen these postulates? Because they are the
    -- simplest ones I could find that enabled me to prove that ≈∋ is
    -- an equivalence relation. Another option would be to postulate
    -- castVarEq + symVar + transVar, but having two almost identical
    -- axioms seems preferable to having three clearly different ones.

    castVarEqˡ (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (v¹ : Γ¹ ∋ τ¹)
               (Γ² : Ctxt) (τ² : Ty Γ²) (v² : Γ² ∋ τ²)
                           (τ³ : Ty Γ¹)
      : (eqτ : τ¹ ≛ τ³) -> v¹ ≈∋ v² -> v¹ ∷≡∋ eqτ ≈∋ v²
    castVarEqʳ (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (v¹ : Γ¹ ∋ τ¹)
               (Γ² : Ctxt) (τ² : Ty Γ²) (v² : Γ² ∋ τ²)
                           (τ³ : Ty Γ²)
      : (eqτ : τ² ≛ τ³) -> v¹ ≈∋ v² -> v¹ ≈∋ v² ∷≡∋ eqτ

  -- Term equality. (The least congruence on terms such that the cast,
  -- beta and eta laws below hold.)

  (≈⊢) (Γ¹, Γ² : Ctxt) (τ¹ : Ty Γ¹) (τ² : Ty Γ²)
    : Γ¹ ⊢ τ¹ -> Γ² ⊢ τ² -> Set
  (≈⊢) = TmEq _ _ _ _

  data TmEq :  (Γ¹, Γ² : Ctxt) -> (τ¹ : Ty Γ¹) -> (τ² : Ty Γ²)
            -> Γ¹ ⊢ τ¹ -> Γ² ⊢ τ² -> Set where

    -- Equivalence.

    reflTm (Γ : Ctxt) (τ : Ty Γ)
      : (t : Γ ⊢ τ) -> t ≈⊢ t
    symTm (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹)
          (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²)
      : t¹ ≈⊢ t² -> t² ≈⊢ t¹
    transTm (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹)
            (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²)
            (Γ³ : Ctxt) (τ³ : Ty Γ³) (t³ : Γ³ ⊢ τ³)
      : t¹ ≈⊢ t² -> t² ≈⊢ t³ -> t¹ ≈⊢ t³

    -- Congruence (castTmCong proved below).

    varCong (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (v¹ : Γ¹ ∋ τ¹)
            (Γ² : Ctxt) (τ² : Ty Γ²) (v² : Γ² ∋ τ²)
      : v¹ ≈∋ v² -> var v¹ ≈⊢ var v²
    λCong (Γ¹ : Ctxt) (τ₁¹ : Ty Γ¹) (τ₂¹ : Ty (Γ¹ ▻ τ₁¹))
          (t¹ : Γ¹ ▻ τ₁¹ ⊢ τ₂¹)
          (Γ² : Ctxt) (τ₁² : Ty Γ²) (τ₂² : Ty (Γ² ▻ τ₁²))
          (t² : Γ² ▻ τ₁² ⊢ τ₂²)
      : t¹ ≈⊢ t² -> λ t¹ ≈⊢ λ t²
    appCong (Γ¹ : Ctxt) (τ₁¹ : Ty Γ¹) (τ₂¹ : Ty (Γ¹ ▻ τ₁¹))
            (t₁¹ : Γ¹ ⊢ Π τ₁¹ τ₂¹) (t₂¹ : Γ¹ ⊢ τ₁¹)
            (Γ² : Ctxt) (τ₁² : Ty Γ²) (τ₂² : Ty (Γ² ▻ τ₁²))
            (t₁² : Γ² ⊢ Π τ₁² τ₂²) (t₂² : Γ² ⊢ τ₁²)
      : t₁¹ ≈⊢ t₁² -> t₂¹ ≈⊢ t₂² -> t₁¹ @ t₂¹ ≈⊢ t₁² @ t₂²
    substTmCong
      (Γ¹, Δ¹ : Ctxt) (ρ¹ : Subst Γ¹ Δ¹) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹)
      (Γ², Δ² : Ctxt) (ρ² : Subst Γ² Δ²) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²)
      : t¹ ≈⊢ t² -> ρ¹ ≘ ρ² -> t¹ /⊢ ρ¹ ≈⊢ t² /⊢ ρ²

    -- Casts.

    castTmEq (Γ : Ctxt) (τ¹, τ² : Ty Γ)
      :  (t : Γ ⊢ τ¹) -> (eqτ : τ¹ ≛ τ²)
      -> t ∷≡⊢ eqτ ≈⊢ t

    -- Evaluation.

    β (Γ : Ctxt) (τ₁ : Ty Γ) (τ₂ : Ty (Γ ▻ τ₁))
      :  (t₁ : Γ ▻ τ₁ ⊢ τ₂) -> (t₂ : Γ ⊢ τ₁)
      -> λ t₁ @ t₂ ≈⊢ t₁ /⊢ sub t₂

    -- η equality.

    η (Γ : Ctxt) (τ₁ : Ty Γ) (τ₂ : Ty (Γ ▻ τ₁))
      :  (t : Γ ⊢ Π τ₁ τ₂)
      -> λ ((t /⊢ wk τ₁) @ var (vz τ₁)) ≈⊢ t

    -- Substitution application axioms.

    substWk (Γ : Ctxt) (τ : Ty Γ)
      :  (v : Γ ∋ τ) -> (σ : Ty Γ)
      -> var v /⊢ wk σ ≈⊢ var (vs v σ)

    substVzSub (Γ : Ctxt)
      :  (σ : Ty Γ) -> (t : Γ ⊢ σ)
      -> var (vz σ) /⊢ sub t ≈⊢ t

    substVsSub (Γ : Ctxt) (τ : Ty Γ)
      :  (v : Γ ∋ τ) -> (σ : Ty Γ) -> (t : Γ ⊢ σ)
      -> var (vs v σ) /⊢ sub t ≈⊢ var v

    substVzLift (Γ, Δ : Ctxt)
      :  (σ : Ty Γ) -> (ρ : Subst Γ Δ)
      -> var (vz σ) /⊢ (ρ ↑ σ) ≈⊢ var (vz $ σ / ρ)

    substVsLift (Γ, Δ : Ctxt) (τ : Ty Γ)
      :  (v : Γ ∋ τ) -> (σ : Ty Γ) -> (ρ : Subst Γ Δ)
      -> var (vs v σ) /⊢ (ρ ↑ σ) ≈⊢ var v /⊢ ρ /⊢ wk (σ / ρ)

    substλ (Γ, Δ : Ctxt) (τ₁ : Ty Γ) (τ₂ : Ty (Γ ▻ τ₁))
      :  (t : Γ ▻ τ₁ ⊢ τ₂) -> (ρ : Subst Γ Δ)
      -> λ t /⊢ ρ ≈⊢ λ (t /⊢ ρ ↑ τ₁)

    substApp (Γ, Δ : Ctxt) (τ₁ : Ty Γ) (τ₂ : Ty (Γ ▻ τ₁))
      :  (t₁ : Γ ⊢ Π τ₁ τ₂) -> (t₂ : Γ ⊢ τ₁) -> (ρ : Subst Γ Δ)
      -> (t₁ @ t₂) /⊢ ρ ≈⊢ (t₁ /⊢ ρ) @ (t₂ /⊢ ρ)

    idVanishesTm (Γ : Ctxt) (τ : Ty Γ)
      : (t : Γ ⊢ τ) -> t /⊢ id Γ ≈⊢ t

    compSplitsTm (Γ, Δ, X : Ctxt) (τ : Ty Γ)
      :  (t : Γ ⊢ τ) -> (ρ₁ : Subst Γ Δ) -> (ρ₂ : Subst Δ X)
      -> t /⊢ ρ₁ ⊙ ρ₂ ≈⊢ t /⊢ ρ₁ /⊢ ρ₂

------------------------------------------------------------------------
-- Right contexts

-- RCtxt Γ is a context extension which can be put "to the right of"
-- the context Γ by using ++.

mutual

  data RCtxt : Ctxt -> Set where
    εʳ              : (Γ : Ctxt) -> RCtxt Γ
    (▻ʳ) (Γ : Ctxt) : (Γ' : RCtxt Γ) -> Ty (Γ ++ Γ') -> RCtxt Γ

  (++) : (Γ : Ctxt) -> RCtxt Γ -> Ctxt
  Γ ++ (εʳ Γ)    = Γ
  Γ ++ (Γ' ▻ʳ τ) = (Γ ++ Γ') ▻ τ

mutual

  -- Right context equality.

  (≜→) (Γ¹, Γ² : Ctxt) : RCtxt Γ¹ -> RCtxt Γ² -> Set
  (≜→) = RCtxtEq _ _

  data RCtxtEq : (Γ¹, Γ² : Ctxt) -> RCtxt Γ¹ -> RCtxt Γ² -> Set where

    -- Equivalence relation axioms proved below.

    εʳCong (Γ¹, Γ² : Ctxt)
      : (eqΓ : Γ¹ ≜ Γ²) -> εʳ Γ¹ ≜→ εʳ Γ²
    (▻ʳCong) (Γ¹ : Ctxt) (Γ'¹ : RCtxt Γ¹) (τ¹ : Ty (Γ¹ ++ Γ'¹))
             (Γ² : Ctxt) (Γ'² : RCtxt Γ²) (τ² : Ty (Γ² ++ Γ'²))
      : Γ'¹ ≜→ Γ'² -> τ¹ ≛ τ² -> Γ'¹ ▻ʳ τ¹ ≜→ Γ'² ▻ʳ τ²
